/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * cache.c,
 * Caching for server side for UDP/IP based RPC.
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <rpc/rpc.h>

#include "rpc_mac.h"
#include "cache.h"


/*
 * Fifo cache for udp server
 * Copies pointers to reply buffers into fifo cache
 * Buffers are sent again if retransmissions are detected.
 */

#define SPARSENESS         4  /* 75% sparse */
#define CACHE_PERROR(msg)  MSG (msg)
#define ALLOC(type, size)  (type*) malloc ((sizeof(type)*(size)))
#define BZERO(a,t,s)       memset ((void*)a,0,sizeof(t)*s)
#define rpc_buffer(xprt)   (void*)((xprt)->xp_p1)

/*
 * An entry in the cache
 */
typedef struct cache_node *cache_ptr;

struct cache_node {
       /* Index into cache is xid, proc, vers, prog and address */
       u_long             cache_xid;
       u_long             cache_proc;
       u_long             cache_vers;
       u_long             cache_prog;
       struct sockaddr_in cache_addr;

       /* The cached reply and length */
       char     *cache_reply;
       u_long    cache_replylen;

       /* Next node on the list, if there is a collision */
       cache_ptr cache_next;
     };

/*
 * The entire cache
 */
struct udp_cache {
  u_long             uc_size;       /* size of cache */
  cache_ptr         *uc_entries;    /* hash table of entries in cache */
  cache_ptr         *uc_fifo;       /* fifo list of entries in cache */
  u_long             uc_nextvictim; /* points to next victim in fifo list */
  u_long             uc_prog;       /* saved program number */
  u_long             uc_vers;       /* saved version number */
  u_long             uc_proc;       /* saved procedure number */
  struct sockaddr_in uc_addr;       /* saved caller's address */
};


/*
 * the hashing function
 */
#define CACHE_LOC(transp, xid)  \
        (xid % (SPARSENESS*((struct udp_cache*)su_data(transp)->su_cache)->uc_size))

/*
 * Enable use of the cache. 
 * Note: there is no disable.
 */
int svcudp_enablecache (SVCXPRT *transp, u_long size)
{
  struct svcudp_data *su = su_data(transp);
  struct udp_cache   *uc;

  if (su->su_cache)
  {
    CACHE_PERROR ("enablecache: cache already enabled");
    return (0);
  }
  uc = ALLOC (struct udp_cache, 1);
  if (uc == NULL)
  {
    CACHE_PERROR ("enablecache: could not allocate cache");
    return (0);
  }
  uc->uc_size       = size;
  uc->uc_nextvictim = 0;
  uc->uc_entries    = ALLOC (cache_ptr, size * SPARSENESS);
  if (uc->uc_entries == NULL)
  {
    CACHE_PERROR ("enablecache: could not allocate cache data");
    return (0);
  }
  BZERO (uc->uc_entries, cache_ptr, size * SPARSENESS);
  uc->uc_fifo = ALLOC (cache_ptr, size);
  if (uc->uc_fifo == NULL)
  {
    CACHE_PERROR ("enablecache: could not allocate cache fifo");
    return (0);
  }
  BZERO (uc->uc_fifo, cache_ptr, size);
  su->su_cache = (char *) uc;
  return (1);
}


/*
 * Set an entry in the cache
 */
void cache_set (SVCXPRT *xprt, u_long replylen)
{
  cache_ptr  victim;
  cache_ptr *vicp;
  u_int      loc;
  char      *newbuf;

  struct svcudp_data *su = su_data (xprt);
  struct udp_cache   *uc = (struct udp_cache *) su->su_cache;

  /*
   * Find space for the new entry, either by
   * reusing an old entry, or by mallocing a new one
   */
  victim = uc->uc_fifo [uc->uc_nextvictim];
  if (victim)
  {
    loc = CACHE_LOC (xprt, victim->cache_xid);
    for (vicp = &uc->uc_entries[loc]; 
         *vicp && *vicp != victim;
         vicp = &(*vicp)->cache_next)
        ;
    if (*vicp == NULL)
    {
      CACHE_PERROR ("cache_set: victim not found");
      return;
    }
    *vicp  = victim->cache_next;  /* remote from cache */
    newbuf = victim->cache_reply;
  }
  else
  {
    victim = ALLOC (struct cache_node, 1);
    if (victim == NULL)
    {
      CACHE_PERROR ("cache_set: victim alloc failed");
      return;
    }
    newbuf = malloc (su->su_iosz);
    if (newbuf == NULL)
    {
      CACHE_PERROR ("cache_set: could not allocate new rpc_buffer");
      return;
    }
  }

  /*
   * Store it away
   */
  victim->cache_replylen = replylen;
  victim->cache_reply    = rpc_buffer (xprt);
  rpc_buffer (xprt) = newbuf;
  xdrmem_create (&su->su_xdrs, rpc_buffer(xprt), su->su_iosz, XDR_ENCODE);
  victim->cache_xid  = su->su_xid;
  victim->cache_proc = uc->uc_proc;
  victim->cache_vers = uc->uc_vers;
  victim->cache_prog = uc->uc_prog;
  victim->cache_addr = uc->uc_addr;
  loc = CACHE_LOC (xprt, victim->cache_xid);
  victim->cache_next   = uc->uc_entries [loc];
  uc->uc_entries [loc] = victim;
  uc->uc_fifo [uc->uc_nextvictim++] = victim;
  uc->uc_nextvictim %= uc->uc_size;
}

/*
 * Try to get an entry from the cache
 * return 1 if found, 0 if not found
 */
int cache_get (SVCXPRT *xprt, struct rpc_msg *msg, char **replyp, u_long *replylenp)
{
  u_int               loc;
  cache_ptr           ent;
  struct svcudp_data *su = su_data (xprt);
  struct udp_cache   *uc = (struct udp_cache*) su->su_cache;

#define EQADDR(a1,a2)  (!memcmp((const void*)&a1,(const void*)&a2,sizeof(a1)))

  loc = CACHE_LOC (xprt, su->su_xid);
  for (ent = uc->uc_entries[loc]; ent; ent = ent->cache_next)
  {
    if (ent->cache_xid  == su->su_xid  &&
        ent->cache_proc == uc->uc_proc &&
        ent->cache_vers == uc->uc_vers &&
        ent->cache_prog == uc->uc_prog &&
        EQADDR(ent->cache_addr, uc->uc_addr))
    {
      *replyp    = ent->cache_reply;
      *replylenp = ent->cache_replylen;
      return (1);
    }
  }
  /*
   * Failed to find entry
   * Remember a few things so we can do a set later
   */
  uc->uc_proc = msg->rm_call.cb_proc;
  uc->uc_vers = msg->rm_call.cb_vers;
  uc->uc_prog = msg->rm_call.cb_prog;
  uc->uc_addr = xprt->xp_raddr;
  return (0);
}
