/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)svc_udp.c 1.24 87/08/11 Copyr 1984 Sun Micro";
#endif

/*
 * svc_udp.c,
 * Server side for UDP/IP based RPC.  (Does some caching in the hopes of
 * achieving execute-at-most-once semantics.)
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <rpc/rpc.h>
#include <sys/socket.h>

#include "cache.h"
#include "rpc_mac.h"

#define rpc_buffer(xprt)  (void*)((xprt)->xp_p1)

static bool_t         svcudp_recv();
static bool_t         svcudp_reply();
static enum xprt_stat svcudp_stat();
static bool_t         svcudp_getargs();
static bool_t         svcudp_freeargs();
static void           svcudp_destroy();

static struct xp_ops svcudp_op = {
       svcudp_recv,
       svcudp_stat,
       svcudp_getargs,
       svcudp_reply,
       svcudp_freeargs,
       svcudp_destroy
     };

/*
 * Usage:
 *  xprt = svcudp_create(sock);
 *
 * If sock<0 then a socket is created, else sock is used.
 * If the socket, sock is not bound to a port then svcudp_create
 * binds it to an arbitrary port.  In any (successful) case,
 * xprt->xp_sock is the registered socket number and xprt->xp_port is the
 * associated port number.
 * Once *xprt is initialized, it is registered as a transporter;
 * see (svc.h, xprt_register).
 * The routines returns NULL if a problem occurred.
 */
SVCXPRT * svcudp_bufcreate (int sock, u_int sendsz, u_int recvsz)
{
  bool_t              madesock = FALSE;
  SVCXPRT            *xprt;
  struct svcudp_data *su;
  struct sockaddr_in addr;
  int    len = sizeof (struct sockaddr_in);

  if (sock == RPC_ANYSOCK)
  {
    if ((sock = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP)) < 0)
    {
      perror ("svcudp_create: socket creation problem");
      return (NULL);
    }
    madesock = TRUE;
  }
  memset (&addr,0,sizeof (addr));
  addr.sin_family = AF_INET;
  if (bindresvport(sock, &addr))
  {
    addr.sin_port = 0;
    bind (sock, (struct sockaddr *)&addr, len);
  }
  if (getsockname(sock,(struct sockaddr*)&addr,&len))
  {
    perror ("svcudp_create - cannot getsockname");
    if (madesock)
       close_s (sock);
    return (NULL);
  }
  xprt = malloc (sizeof(SVCXPRT));
  if (!xprt)
  {
    MSG ("svcudp_create: out of memory");
    return (NULL);
  }
  su = malloc (sizeof(*su));
  if (!su)
  {
    MSG ("svcudp_create: out of memory");
    return (NULL);
  }
  su->su_iosz = ((MAX(sendsz, recvsz) + 3) / 4) * 4;
  if ((rpc_buffer(xprt) = malloc (su->su_iosz)) == NULL)
  {
    MSG ("svcudp_create: out of memory");
    return (NULL);
  }
  xdrmem_create (&su->su_xdrs, rpc_buffer(xprt), su->su_iosz, XDR_DECODE);
  su->su_cache = NULL;
  xprt->xp_p2           = (caddr_t) su;
  xprt->xp_verf.oa_base = (caddr_t) su->su_verfbody;
  xprt->xp_ops          = &svcudp_op;
  xprt->xp_port         = ntohs (addr.sin_port);
  xprt->xp_sock         = sock;
  xprt_register (xprt);
  return (xprt);
}

SVCXPRT *svcudp_create (int sock)
{
  return svcudp_bufcreate (sock, UDPMSGSIZE, UDPMSGSIZE);
}

static enum xprt_stat svcudp_stat (SVCXPRT *xprt)
{
  ARGSUSED (xprt);
  return (XPRT_IDLE);
}

static bool_t svcudp_recv (SVCXPRT *xprt, struct rpc_msg *msg)
{
  struct svcudp_data *su = su_data (xprt);
  XDR   *xdrs = &su->su_xdrs;
  int    rlen;
  char  *reply;
  u_long replylen;

again:
  xprt->xp_addrlen = sizeof (struct sockaddr_in);
  rlen = recvfrom (xprt->xp_sock, rpc_buffer(xprt), su->su_iosz,
                   0, (struct sockaddr*) &xprt->xp_raddr, &xprt->xp_addrlen);

  if (rlen == -1 && errno == EINTR)
     goto again;

  if (rlen < 4*sizeof(u_long))
     return (FALSE);

  xdrs->x_op = XDR_DECODE;
  XDR_SETPOS (xdrs, 0);
  if (!xdr_callmsg(xdrs, msg))
    return (FALSE);

  su->su_xid = msg->rm_xid;
  if (su->su_cache && cache_get(xprt, msg, &reply, &replylen))
     sendto (xprt->xp_sock, reply, (int)replylen, 0,
             (struct sockaddr*)&xprt->xp_raddr, xprt->xp_addrlen);

  return (TRUE);
}

static bool_t svcudp_reply (SVCXPRT *xprt, struct rpc_msg *msg)
{
  struct svcudp_data *su   = su_data (xprt);
  XDR                *xdrs = &su->su_xdrs;
  int    slen;
  bool_t stat = FALSE;

  xdrs->x_op = XDR_ENCODE;
  XDR_SETPOS (xdrs, 0);
  msg->rm_xid = su->su_xid;

  if (xdr_replymsg(xdrs, msg))
  {
    slen = (int)XDR_GETPOS (xdrs);
    if (sendto(xprt->xp_sock, rpc_buffer(xprt), slen, 0,
               (struct sockaddr*) &xprt->xp_raddr, xprt->xp_addrlen) == slen)
    {
      stat = TRUE;
      if (su->su_cache && slen >= 0)
         cache_set (xprt, (u_long) slen);
    }
  }
  return (stat);
}

static bool_t svcudp_getargs (SVCXPRT *xprt, xdrproc_t xdr_args, caddr_t args_ptr)
{
  return (*xdr_args) (&(su_data(xprt)->su_xdrs), args_ptr);
}

static bool_t svcudp_freeargs (SVCXPRT *xprt, xdrproc_t xdr_args, caddr_t args_ptr)
{
  XDR *xdrs = &(su_data(xprt)->su_xdrs);

  xdrs->x_op = XDR_FREE;
  return (*xdr_args) (xdrs, args_ptr);
}

static void svcudp_destroy (SVCXPRT *xprt)
{
  struct svcudp_data *su = su_data(xprt);

  xprt_unregister (xprt);
  close_s (xprt->xp_sock);
  XDR_DESTROY (&su->su_xdrs);
  free (rpc_buffer(xprt));
  free (su);
  free (xprt);
}

